use std::num::Wrapping;

use crate::Hasher::HasherTrait;


#[derive(Debug, Clone)]
pub struct Sha1
{
	h0: Wrapping<u32>,
	h1: Wrapping<u32>,
	h2: Wrapping<u32>,
	h3: Wrapping<u32>,
	h4: Wrapping<u32>,
	leftoverInput: Vec<u8>,
	inputLen: usize,
}

impl Sha1
{
	pub fn New () -> Self
	{
		Self
		{
			h0: Wrapping(0x67452301),
			h1: Wrapping(0xEFCDAB89),
			h2: Wrapping(0x98BADCFE),
			h3: Wrapping(0x10325476),
			h4: Wrapping(0xC3D2E1F0),
			leftoverInput: Vec::new (),
			inputLen: 0,
		}
	}


	fn AddChunk (&mut self, input: &[u8; 64])
	{
		let mut w = [0u32; 80];

		// make w[0..16]
		for i in (0..input.len ()).step_by (4)
		{
			w[i/4] = u32::from_be_bytes (input[i..i + 4].try_into ().unwrap ());
		}

		// make w[16..80]
		for i in 16..80
		{
			w[i] = (w[i - 3] ^ w[i - 8] ^ w[i - 14] ^ w[i - 16]).rotate_left (1);
		}

		// init values
		let mut a = self.h0;
		let mut b = self.h1;
		let mut c = self.h2;
		let mut d = self.h3;
		let mut e = self.h4;

		// do 80 rounds
		for i in 0..80
		{
			let (f, k) = match i
			{
				x if (0..20).contains (&x) =>
				{
					((b & c) | ((!b) & d), Wrapping(0x5A827999))
				},
				x if (20..40).contains (&x) =>
				{
					(b ^ c ^ d, Wrapping(0x6ED9EBA1))
				},
				x if (40..60).contains (&x) =>
				{
					((b & c) | (b & d) | (c & d), Wrapping(0x8F1BBCDC))
				},
				x if (60..80).contains (&x) =>
				{
					(b ^ c ^ d, Wrapping(0xCA62C1D6))
				},
				_ => panic! ("Something is wrong"),
			};


			let temp = a.rotate_left (5) + f + e + k + Wrapping(w[i]);
			e = d;
			d = c;
			c = b.rotate_left (30);
			b = a;
			a = temp;
		}


		// add this chunk hash to current hash
		self.h0 += a;
		self.h1 += b;
		self.h2 += c;
		self.h3 += d;
		self.h4 += e;
	}

	pub fn AddInput (&mut self, input: &[u8])
	{
		if self.leftoverInput.len () + input.len () < 64
		{
			// not enough byte to process
			self.leftoverInput.extend (input);

			// update input len
			self.inputLen += input.len ();
			return;
		}

		// first chunk
		let buffer = [&self.leftoverInput, &input[0..64 - self.leftoverInput.len ()]].concat ();
		assert! (buffer.len () == 64);
		self.AddChunk (&buffer[0..64].try_into ().unwrap ());

		// do next n chunks
		for chunk in input[64 - self.leftoverInput.len ()..].chunks_exact (64)
		{
			self.AddChunk (chunk.try_into ().unwrap ());
		}

		// push last chunk into leftover

		// .......................................
		// |      |      |      |      |      |

		// .......................................
		//     |      |      |      |      |      |

		// .......................................
		//   |      |      |      |      |      |
		let x = (input.len () / 64)*64;
		let mut y = x + (64 - self.leftoverInput.len ());

		if y > input.len ()
		{
			y -= 64;
		}

		self.leftoverInput.clear ();
		self.leftoverInput.extend (&input[y..]);

		// update input len
		self.inputLen += input.len ();
	}

	pub fn GetOutput (mut self) -> [u8; Self::OUTPUT_SIZE]
	{
		let paddingLen = ((self.inputLen + 1 + 8 + 64 - 1)/64)*64 - (self.inputLen + 1 + 8);
		let lastBlock = [&[0x80][..], &vec! [0x00; paddingLen], &(self.inputLen*8).to_be_bytes ()].concat ();
		self.AddInput (&lastBlock);

		[self.h0.0.to_be_bytes (), self.h1.0.to_be_bytes (), self.h2.0.to_be_bytes (), self.h3.0.to_be_bytes (), self.h4.0.to_be_bytes ()].concat ().try_into ().unwrap ()
	}

	pub fn Hash (input: &[u8]) -> [u8; Self::OUTPUT_SIZE]
	{
		let mut hasher = Self::New ();
		hasher.AddInput (input);
		hasher.GetOutput ()
	}
}


impl HasherTrait for Sha1
{
	const OUTPUT_SIZE: usize = 20;

	fn AddInput (&mut self, input: &[u8])
	{
		self.AddInput (input);
	}

	fn GetOutput (self) -> [u8; Self::OUTPUT_SIZE]
	{
		self.GetOutput ()
	}

	fn Hash (input: &[u8]) -> [u8; Self::OUTPUT_SIZE]
	{
		Self::Hash (input)
	}
}


pub struct Sha2_256;
pub struct Sha2_384;
pub struct Sha2_512;
pub struct Sha3_224;
pub struct Sha3_256;
pub struct Sha3_384;
pub struct Sha3_512;


#[cfg(test)]
mod tests
{
	use super::*;
	use hex;

	#[test]
	fn Sha1_1 ()
	{
		let sha1 = Sha1::New ();
		let output = sha1.GetOutput ();
		assert! (hex::encode (output) == "da39a3ee5e6b4b0d3255bfef95601890afd80709");
	}

	#[test]
	fn Sha1_2 ()
	{
		let mut sha1 = Sha1::New ();
		sha1.AddInput ("abc".as_bytes ());
		let output = sha1.GetOutput ();
		assert! (hex::encode (output) == "a9993e364706816aba3e25717850c26c9cd0d89d");
	}

	#[test]
	fn Sha1_3 ()
	{
		let mut sha1 = Sha1::New ();
		sha1.AddInput ("abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq".as_bytes ());
		let output = sha1.GetOutput ();
		assert! (hex::encode (output) == "84983e441c3bd26ebaae4aa1f95129e5e54670f1");
	}
}

